@@ -11,6 +11,7 @@ from playwright.async_api import (
     Page,
     PlaywrightContextManager,
     Request as PlaywrightRequest,
+    Response as PlaywrightResponse,
     Route,
 )
 from scrapy import Spider, signals
@@ -35,6 +36,26 @@ PlaywrightHandler = TypeVar("PlaywrightHandler", bound="ScrapyPlaywrightDownload
 logger = logging.getLogger("scrapy-playwright")
 
 
+def _make_request_logger(context_name: str) -> Callable:
+    def _log_request(request: PlaywrightRequest) -> None:
+        logger.debug(
+            f"[Context={context_name}] Request: <{request.method.upper()} {request.url}> "
+            f"(resource type: {request.resource_type}, referrer: {request.headers.get('referer')})"
+        )
+
+    return _log_request
+
+
+def _make_response_logger(context_name: str) -> Callable:
+    def _log_request(response: PlaywrightResponse) -> None:
+        logger.debug(
+            f"[Context={context_name}] Response: <{response.status} {response.url}> "
+            f"(referrer: {response.headers.get('referer')})"
+        )
+
+    return _log_request
+
+
 class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     def __init__(self, crawler: Crawler) -> None:
         super().__init__(settings=crawler.settings, crawler=crawler)
@@ -107,6 +128,9 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             context = await self._create_browser_context(context_name, context_kwargs)
             self.contexts[context_name] = context
         page = await context.new_page()
+        page.on("response", _make_response_logger(context_name))
+        page.on("request", _make_request_logger(context_name))
+        page.on("request", self._increment_request_stats)
         self.stats.inc_value("playwright/page_count")
         if self.default_navigation_timeout:
             page.set_default_navigation_timeout(self.default_navigation_timeout)
@@ -115,12 +139,14 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     @inlineCallbacks
     def close(self) -> Deferred:
         yield super().close()
-        for context in self.contexts.copy().values():
-            yield deferred_from_coro(context.close())
+        yield deferred_from_coro(self._close())
+
+    async def _close(self) -> None:
+        self.contexts.clear()
         if getattr(self, "browser", None):
             logger.info("Closing browser")
-            yield deferred_from_coro(self.browser.close())
-        yield deferred_from_coro(self.playwright_context_manager.__aexit__())
+            await self.browser.close()
+        await self.playwright_context_manager.__aexit__()
 
     def download_request(self, request: Request, spider: Spider) -> Deferred:
         if request.meta.get("playwright"):
@@ -131,6 +157,21 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         page = request.meta.get("playwright_page")
         if not isinstance(page, Page):
             page = await self._create_page(request)
+
+        # attach event handlers
+        event_handlers = request.meta.get("playwright_page_event_handlers") or {}
+        for event, handler in event_handlers.items():
+            if callable(handler):
+                page.on(event, handler)
+            elif isinstance(handler, str):
+                try:
+                    page.on(event, getattr(spider, handler))
+                except AttributeError:
+                    logger.warning(
+                        f"Spider '{spider.name}' does not have a '{handler}' attribute,"
+                        f" ignoring handler for event '{event}'"
+                    )
+
         await page.unroute("**")
         await page.route(
             "**",
@@ -187,6 +228,14 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             flags=["playwright"],
         )
 
+    def _increment_request_stats(self, request: PlaywrightRequest) -> None:
+        stats_prefix = "playwright/request_count"
+        self.stats.inc_value(stats_prefix)
+        self.stats.inc_value(f"{stats_prefix}/resource_type/{request.resource_type}")
+        self.stats.inc_value(f"{stats_prefix}/method/{request.method}")
+        if request.is_navigation_request():
+            self.stats.inc_value(f"{stats_prefix}/navigation")
+
     def _make_close_browser_context_callback(self, name: str) -> Callable:
         def close_browser_context_callback() -> None:
             logger.debug("Browser context closed: '%s'", name)
@@ -200,12 +249,13 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     ) -> Callable:
         def request_handler(route: Route, pw_request: PlaywrightRequest) -> None:
             """Override request headers, method and body."""
+            headers.setdefault("user-agent", pw_request.headers.get("user-agent"))
             if pw_request.url == url:
                 overrides: dict = {"method": method, "headers": headers}
                 if body is not None:
                     overrides["post_data"] = body.decode(encoding)
-                # otherwise this fails with playwright.helper.Error: NS_ERROR_NET_RESET
                 if self.browser_type == "firefox":
+                    # otherwise this fails with playwright.helper.Error: NS_ERROR_NET_RESET
                     overrides["headers"]["host"] = urlparse(pw_request.url).netloc
             else:
                 overrides = {"headers": pw_request.headers.copy()}
@@ -213,12 +263,5 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
                 if headers.get("user-agent"):
                     overrides["headers"]["user-agent"] = headers["user-agent"]
             asyncio.create_task(route.continue_(**overrides))
-            # increment stats
-            self.stats.inc_value("playwright/request_count")
-            resource_type = pw_request.resource_type
-            self.stats.inc_value(f"playwright/request_count/resource_type/{resource_type}")
-            self.stats.inc_value(f"playwright/request_count/method/{pw_request.method}")
-            if pw_request.is_navigation_request():
-                self.stats.inc_value("playwright/request_count/navigation")
 
         return request_handler
